import functools

from packaging.version import parse as parse_version
from rich import box
from rich.console import Group
from rich.highlighter import RegexHighlighter, ReprHighlighter
from rich.padding import Padding
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from leak import config, logger, parser, rprint


class SomeEmailsHighlighter(RegexHighlighter):
    """Apply style to the emails matching this simple regex"""

    base_style = "repr."
    highlights = [r"(?P<path>[\w\.\+-]+@([\w-]+\.)+[\w-]+)"]


def show_package_info(data):
    info = data["info"]
    name = info["name"]
    versions_count = len(data["releases"])

    highlight_link = ReprHighlighter()
    highlight_email = SomeEmailsHighlighter()
    table = Table(show_header=False, show_footer=False, box=box.SIMPLE, expand=True)
    table.add_row("Author:", parser.get_author(info))
    table.add_row("Email:", highlight_email(parser.get_email(info)))
    table.add_row("Home page:", highlight_link(parser.get_homepage(info)))
    table.add_row("License:", parser.get_license(info))
    table.add_row("Version:", info["version"])

    summary = Padding(Text(f"{info['summary']}", style="bold bright_white"), (1, 2))
    group = Group(
        summary,
        table,
    )
    panel = Panel(
        group,
        expand=True,
        title=f"{name}",
        subtitle=f"{versions_count} versions available",
        border_style="cyan",
        width=config.PANEL_WIDTH,
    )
    rprint(panel)


def show_package_versions(releases, downloads: dict, showall: bool = False):
    most_popular_count = 0  # noqa
    most_popular_release = None  # noqa
    most_recent_release = None
    most_recent_date = config.EPOCH_BEGIN

    try:
        versions = sorted(releases.keys(), reverse=True, key=parse_version)
    except ValueError as e:
        logger.debug(e)
        logger.debug("Trying to sort versions as strings")
        splitter = functools.partial(parser.versions_split, type_applyer=str)
        versions = sorted(releases.keys(), reverse=True, key=splitter)

    for release_num, release_data in releases.items():
        downloads_count = parser.get_max_downloads_for_release(release_data)
        if downloads_count:
            print(downloads_count)

        upload_date = parser.get_latest_time_for_release(release_data)
        if upload_date > most_recent_date:
            most_recent_date = upload_date
            most_recent_release = release_num  # noqa

    table = Table(
        show_header=True,
        show_footer=False,
        box=None,
        expand=True,
    )
    table.add_column(Padding("version", (1, 0)))
    table.add_column(Padding("date", (1, 0)), justify="center")
    if downloads:
        table.add_column(Padding("downloads", (1, 0)), justify="right")

    list_versions = versions if showall else versions[: config.SHOW_LATEST_RELEASES]
    for version in list_versions:
        release_data = releases[version]
        downloads_count = parser.get_downloads_for_version(
            version, downloads_data=downloads
        )
        upload_date = parser.get_latest_time_for_release(release_data)
        upload_date_str = upload_date.strftime(config.DATE_FORMAT)
        row = [version, upload_date_str]
        if downloads:
            row.append(f"{downloads_count:,}")
        table.add_row(*row)

    panel = Panel(
        table,
        expand=True,
        border_style="yellow",
        width=config.PANEL_WIDTH,
        title="List of releases" if showall else "Recent releases",
    )

    rprint(panel)


def warning(message):
    panel = Panel(
        message,
        expand=True,
        border_style="bright_red",
        width=config.PANEL_WIDTH,
        title="▲",
        title_align="left",
    )
    rprint(panel)



import configparser
from datetime import datetime
from pathlib import Path
from typing import Annotated, Any, Optional

from pydantic_settings import (
    BaseSettings,
    EnvSettingsSource,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
)

CONFIG_FILEPATH = Path("~/.config/leak/config.ini").expanduser()
CONFIG_FILEPATH.parent.mkdir(parents=True, exist_ok=True)
DEFAULT_SECTION = "config"


class NotConfgirurable:
    """Annotation to prevent setting of a field value from a configuration file."""

    pass


class ConfigparserSettingsSource(PydanticBaseSettingsSource):
    def get_field_value(self, field, field_name: str) -> tuple[Any, str, bool]:
        NO_VALUE = (None, field_name, False)

        if NotConfgirurable in field.metadata:
            return NO_VALUE
        config_filepath = self.config.get("config_filepath")
        if not config_filepath.exists():
            return NO_VALUE
        config_parser = configparser.ConfigParser()
        config_parser.read(config_filepath)
        if DEFAULT_SECTION not in config_parser:
            return NO_VALUE
        config_field_name = field_name.lower().replace("_", "-")
        field_value = config_parser.get("config", config_field_name, fallback=None)
        return field_value, field_name, False

    def __call__(self) -> dict[str, Any]:
        d: dict[str, Any] = {}

        for field_name, field in self.settings_cls.model_fields.items():
            field_value, field_key, value_is_complex = self.get_field_value(
                field, field_name
            )
            field_value = self.prepare_field_value(
                field_name, field, field_value, value_is_complex
            )
            if field_value is not None:
                d[field_key] = field_value

        return d


class CustomEnvSettingsSource(EnvSettingsSource):
    def get_field_value(self, field, field_name: str) -> tuple[Any, str, bool]:
        if NotConfgirurable in field.metadata:
            return (None, field_name, False)
        return super().get_field_value(field, field_name)


class Settings(BaseSettings):
    model_config = SettingsConfigDict(
        env_prefix="LEAK_",
        case_sensitive=True,
        config_filepath=CONFIG_FILEPATH,
    )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return (
            init_settings,
            ConfigparserSettingsSource(settings_cls),
            CustomEnvSettingsSource(settings_cls),
        )

    @classmethod
    def allowed_config_keys(cls) -> set[str]:
        # TODO: this should also respect `NotConfgirurable` annotation
        normalize_name = lambda name: name.lower().replace("_", "-")  # noqa: E731
        return set(map(normalize_name, cls.model_fields.keys()))

    DEBUG: bool = False
    DATE_FORMAT: str = "%Y/%m/%d %H:%M"
    EPOCH_BEGIN: Annotated[datetime, NotConfgirurable] = datetime.fromtimestamp(0)
    SHOW_LATEST_RELEASES: int = 12
    SHOW_PAGER: int = 30
    PANEL_WIDTH: int = 70
    API_KEY: Optional[str] = None
    SHOW_DOWNLOADS: bool = True


config = Settings()



import logging

from rich.console import Console

from .settings import config

logging.basicConfig(level=logging.CRITICAL)
logger = logging.getLogger(__package__)

if config.DEBUG:
    logger.setLevel(logging.DEBUG)

console = Console(highlight=False)
rprint = console.print



from datetime import datetime
from email.header import decode_header
from email.utils import parseaddr
from typing import Optional

from leak import config, logger


def versions_split(version_str, type_applyer=int) -> list[int]:
    dots_count = version_str.count(".")
    if dots_count == 0:
        major, minor, patch = version_str, 0, 0
    elif dots_count == 1:
        major, minor = version_str.split(".")
        patch = 0
    elif dots_count == 2:
        major, minor, patch = version_str.split(".")
    else:
        logger.debug(
            'Incorrect version "{version}". Move to bottom when sorting'.format(
                version=version_str
            )
        )
        major, minor, patch = 0, 0, 0

    return list(map(type_applyer, (major, minor, patch)))


def get_max_downloads_for_release(release: list[dict]) -> int:
    max_downloads = 0
    for release_record in release:
        if release_record["downloads"] > max_downloads:
            max_downloads = release_record["downloads"]
    return max_downloads


def get_latest_time_for_release(release: list[dict]) -> datetime:
    latest_time = config.EPOCH_BEGIN
    date_format = "%Y-%m-%dT%H:%M:%S"
    for release_record in release:
        upload_time = datetime.strptime(release_record["upload_time"], date_format)
        if upload_time > latest_time:
            latest_time = upload_time
    return latest_time


def get_downloads_for_version(version: str, downloads_data: dict) -> int:
    downloads = 0
    for date, data in downloads_data.items():
        downloads += data.get(version, 0)
    return downloads


def decode_name(encoded_data) -> str:
    decoded_parts = decode_header(encoded_data)
    name, encoding = decoded_parts[0]
    return name.decode(encoding) if isinstance(name, bytes) else name


def get_author(info: dict) -> str:
    if "author" in info and info["author"]:
        return info["author"]

    if "author_email" in info and info["author_email"]:
        emails = info["author_email"].split(",")
        return decode_name(parseaddr(emails[0])[0])

    if "maintainer" in info and info["maintainer"]:
        return info["maintainer"]

    if "maintainer_email" in info and info["maintainer_email"]:
        emails = info["maintainer_email"].split(",")
        return decode_name(parseaddr(emails[0])[0])
    return "n/a"


def get_email(info) -> str:
    if "author_email" in info and info["author_email"]:
        emails = info["author_email"].split(",")
        return parseaddr(emails[0])[1]

    if "maintainer_email" in info and info["maintainer_email"]:
        emails = info["maintainer_email"].split(",")
        return parseaddr(emails[0])[1]
    return "n/a"


def get_homepage(info) -> str:
    if "home_page" in info and info["home_page"]:
        return info["home_page"]
    if "project_url" in info and info["project_url"]:
        return info["project_url"]
    homepage = info.get("project_urls", {}).get("Homepage", "n/a")
    return homepage


def get_license(info) -> str:
    if "license" in info and info["license"]:
        # Return the first non-empty line of the license
        # as license field can contain full license text
        return next(filter(bool, map(str.strip, info["license"].splitlines())))
    if "license_expression" in info and info["license_expression"]:
        return info["license_expression"]
    license_class = get_license_from_classifiers(info.get("classifiers", []))
    return license_class or "n/a"


def get_license_from_classifiers(classifiers: list[str]) -> Optional[str]:
    for classifier in classifiers:
        if classifier.startswith("License"):
            return classifier.split("::")[-1].strip()



import configparser
from typing import Optional

import click

from leak import logger, main, rprint, settings
from leak.ui import warning

config = settings.config


@click.command()
@click.version_option(
    prog_name="leak",
    package_name="leak",
)
@click.argument(
    "package_name",
    required=False,
)
@click.option(
    "-a",
    "--all",
    "showall",
    is_flag=True,
    default=False,
)
@click.option(
    "-s",
    "--set",
    "set_value",
    help="Set a configuration value",
)
@click.pass_context
def cli(
    ctx, package_name: Optional[str], showall: bool, set_value: Optional[str] = None
):
    """Shows all releases for a package and some info about it.

    PACKAGE_NAME Name of the package to fetch info about.
    """
    if package_name is not None and set_value is not None:
        raise click.UsageError(
            "You cannot have configuration set flag together with a package name."
        )

    if set_value is not None:
        logger.debug(f"Config setting invocation: {set_value}")
        set_config_value(set_value)
        return

    if package_name is None:
        raise click.MissingParameter(
            message="You must provide a package name to fetch information about.",
            ctx=ctx,
            param_type="argument",
            param_hint="PACKAGE_NAME",
        )

    main.main(package_name=package_name, showall=showall)

    if config.SHOW_DOWNLOADS and not config.API_KEY:
        warning(
            "API key is not set. Downloads are not shown. Configure it via\n"
            "    [bold red]export LEAK_API_KEY=<your_api_key>[/]\n"
            "or by running the command\n"
            "    [bold red]leak --set api-key=<your_api_key>[/]\n"
            "To simply hide this warning type\n"
            "    [bold red]leak --set show-downloads=false[/]"
        )


def set_config_value(set_value: str):
    """Set a configuration value."""
    if "=" not in set_value:
        raise click.BadParameter(f"Bad format: '{set_value}'. Must be key=value.")

    key, value = set_value.split("=", 1)
    if key not in config.allowed_config_keys():
        raise click.BadParameter(f"Bad configuration key: '{key}'.")

    config_parser = configparser.ConfigParser()
    if settings.CONFIG_FILEPATH.exists():
        config_parser.read(settings.CONFIG_FILEPATH)

    if settings.DEFAULT_SECTION not in config_parser:
        config_parser[settings.DEFAULT_SECTION] = {}

    config_parser[settings.DEFAULT_SECTION][key.strip()] = value.strip()

    with open(settings.CONFIG_FILEPATH, "w") as f:
        config_parser.write(f)

    rprint(f"Configuration value for [bold red]{key}[/] updated.")


if __name__ == "__main__":
    cli()



import sys
from contextlib import contextmanager

import requests

from leak import config, console, logger, ui


@contextmanager
def dummy_context(*args, **kwargs):
    yield


def get_package_data(package_name: str) -> dict:
    url = f"https://pypi.org/pypi/{package_name}/json"
    resp = requests.get(url)
    if resp.status_code != 200:
        raise ValueError("No such package")

    data = resp.json()
    return data


def get_downloads_data(package_name: str) -> dict:
    if not config.API_KEY or not config.SHOW_DOWNLOADS:
        logger.warning("Skipping downloads data retrieval")
        return {}

    url = f"https://api.pepy.tech/api/v2/projects/{package_name}"
    headers = {
        "X-API-Key": config.API_KEY,
    }
    resp = requests.get(url, headers=headers)
    if resp.status_code != 200:
        logger.error(f"Cannot get downloads data: [{resp.status_code}] {resp.text}")
        return {}

    data = resp.json()
    return data.get("downloads", {})


def parse_packages_from_html(html_content):
    return ""


def search_for_package(package_name: str):
    url = f"https://pypi.python.org/pypi?:action=search&term={package_name}"
    resp = requests.get(url)
    if resp.status_code != 200:
        return []
    return parse_packages_from_html(resp.text)


def main(package_name: str = "", showall: bool = False):
    try:
        package_data = get_package_data(package_name)
    except ValueError as e:
        logger.error(e)
        console.print(f"No such package [bold red]{package_name}[/].")
        return sys.exit(1)

    releases = package_data["releases"]
    downloads = get_downloads_data(package_name)
    context = dummy_context
    if showall and len(releases) > config.SHOW_PAGER:
        context = console.pager

    with context(styles=True):
        ui.show_package_info(package_data)
        ui.show_package_versions(releases, downloads, showall)
